/*
 * monotonic.c
 * Copyright (C) 2019 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the GPL3 license.
 */

#define _POSIX_C_SOURCE 200809L
#define MONOTONIC_IMPLEMENTATION
#include "monotonic.h"
