// Unicode data, built from the Unicode Standard 17.0.0
// Code generated by wcwidth.py, DO NOT EDIT.

#pragma once

typedef enum IndicConjunctBreak {
	ICB_None,
	ICB_Linker,
	ICB_Consonant,
	ICB_Extend,
} IndicConjunctBreak;

static const char_type CharProps_mask = 255u;
static const char_type CharProps_shift = 8u;
static const uint8_t CharProps_t1[4352] = {
	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 53, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 54, 52, 52, 52, 55, 21, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 62, 63, 64, 65, 66, 67, 68, 62, 63, 64, 65, 66, 67, 68, 62, 63, 64, 65, 66, 67, 68, 62, 63, 64, 65, 66, 67, 68, 62, 63, 64, 65, 66, 67, 68, 62, 69, 70, 70, 70, 70, 70, 70, 70, 70, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 52, 72, 73, 21, 74, 75, 76, 77, 78, 79, 80, 81, 82, 21, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 21, 21, 21, 108, 109, 110, 111, 111, 111, 111, 111, 111, 111, 111, 111, 112, 21, 21, 21, 21, 113, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 114, 21, 21, 115, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 116, 111, 111, 111, 111, 111, 111, 21, 21, 117, 118, 111, 119, 120, 121, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 122, 123, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 124, 52, 125, 126, 111, 111, 111, 111, 111, 111, 111, 111, 111, 127, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 128, 40, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 40, 40, 139, 111, 111, 111, 111, 140, 141, 142, 143, 111, 144, 145, 146, 147, 148, 149, 111, 111, 150, 151, 152, 111, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 165, 165, 166, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 167, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 168, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 169, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 170, 52, 52, 171, 172, 172, 172, 172, 172, 172, 172, 172, 172, 52, 52, 173, 172, 172, 172, 172, 174, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 175, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 176, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 174, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 178, 179, 180, 180, 180, 180, 180, 180, 180, 180, 180, 180, 180, 180, 180, 180, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 111, 177, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 181, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 71, 181
};
static const uint8_t CharProps_t2[46592] = {
	0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 5, 5, 6, 7, 5, 5, 5, 8, 9, 6, 10, 5, 11, 5, 5, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 5, 5, 10, 10, 10, 5, 5, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 8, 5, 9, 14, 15, 14, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 8, 10, 9, 10, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 17, 7, 7, 18, 7, 19, 17, 20, 21, 22, 23, 10, 24, 25, 14, 26, 27, 28, 28, 20, 16, 17, 17, 20, 28, 22, 29, 28, 28, 28, 17, 13, 13, 13, 13, 13, 13, 30, 13, 13, 13, 13, 13, 13, 13, 13, 13, 30, 13, 13, 13, 13, 13, 13, 27, 30, 13, 13, 13, 13, 13, 30, 31, 31, 31, 16, 16, 16, 16, 31, 16, 31, 31, 31, 16, 31, 31, 16, 16, 31, 16, 31, 31, 16, 16, 16, 27, 31, 31, 31, 16, 31, 16, 31, 16, 13, 31, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 31, 13, 31, 13, 16, 13, 16, 13, 16, 13, 31, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 30, 31, 13, 16, 13, 31, 13, 16, 13, 16, 13, 31, 30, 31, 13, 16, 13, 16, 31, 13, 16, 13, 16, 13, 16, 30, 31, 30, 31, 13, 31, 13, 16, 13, 31, 31, 30, 31, 13, 31, 13, 16, 13, 16, 30, 31, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 30, 31, 13, 16, 13, 31, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 13, 16, 13, 16, 13, 16, 16, 16, 13, 13, 16, 13, 16, 13, 13, 16, 13, 13, 13, 16, 16, 13, 13, 13, 13, 16, 13, 13, 16, 13, 13, 13, 16, 16, 16, 13, 13, 16, 13, 13, 16, 13, 16, 13, 16, 13, 13, 16, 13, 16, 16, 13, 16, 13, 13, 16, 13, 13, 13, 16, 13, 16, 13, 13, 16, 16, 32, 13, 16, 16, 16, 32, 32, 32, 32, 13, 33, 16, 13, 33, 16, 13, 33, 16, 13, 31, 13, 31, 13, 31, 13, 31, 13, 31, 13, 31, 13, 31, 13, 31, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 13, 33, 16, 13, 16, 13, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 16, 16, 16, 16, 16, 13, 13, 16, 13, 13, 16, 16, 13, 16, 13, 13, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 31, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 31, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 32, 32, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 14, 14, 20, 14, 34, 35, 34, 35, 35, 35, 34, 35, 34, 34, 35, 34, 14, 14, 14, 14, 14, 14, 20, 20, 20, 20, 14, 20, 14, 20, 34, 34, 34, 34, 34, 14, 14, 14, 14, 14, 14, 14, 34, 14, 34, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 37, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 13, 16, 13, 16, 34, 14, 13, 16, 38, 38, 34, 16, 16, 16, 5, 13, 38, 38, 38, 38, 14, 14, 13, 5, 13, 13, 13, 38, 13, 38, 13, 13, 16, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 38, 30, 30, 30, 30, 30, 30, 30, 13, 13, 16, 16, 16, 16, 16, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 16, 31, 31, 31, 31, 31, 31, 31, 16, 16, 16, 16, 16, 13, 16, 16, 13, 13, 13, 16, 16, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 16, 16, 16, 13, 16, 10, 13, 16, 13, 13, 16, 16, 13, 13, 13, 13, 30, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 16, 31, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 19, 36, 36, 36, 36, 36, 39, 39, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 34, 5, 5, 5, 5, 5, 5, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 5, 11, 38, 38, 19, 19, 7, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 11, 36, 5, 36, 36, 5, 36, 36, 5, 36, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 32, 32, 32, 32, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 40, 40, 40, 40, 40, 40, 10, 10, 10, 5, 5, 7, 5, 5, 19, 19, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 5, 24, 5, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 5, 5, 5, 32, 32, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 5, 32, 36, 36, 36, 36, 36, 36, 36, 40, 19, 36, 36, 36, 36, 36, 36, 34, 34, 36, 36, 19, 36, 36, 36, 36, 32, 32, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 32, 19, 19, 32, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 38, 40, 32, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 34, 34, 19, 5, 5, 5, 34, 38, 38, 36, 7, 7, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 34, 36, 36, 36, 36, 36, 36, 36, 36, 36, 34, 36, 36, 36, 34, 36, 36, 36, 36, 36, 38, 38, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 38, 38, 5, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 14, 32, 32, 32, 32, 32, 32, 32, 40, 40, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 40, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 42, 36, 32, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 42, 42, 42, 42, 44, 42, 42, 32, 36, 36, 36, 36, 36, 36, 36, 43, 43, 43, 43, 43, 43, 43, 43, 32, 32, 36, 36, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 34, 32, 32, 32, 32, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 32, 36, 42, 42, 38, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 38, 38, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 38, 43, 38, 38, 38, 43, 43, 43, 43, 38, 38, 36, 32, 45, 42, 42, 36, 36, 36, 36, 38, 38, 42, 42, 38, 38, 42, 42, 44, 32, 38, 38, 38, 38, 38, 38, 38, 38, 45, 38, 38, 38, 38, 43, 43, 38, 43, 32, 32, 36, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 43, 43, 7, 7, 46, 46, 46, 46, 46, 46, 19, 7, 32, 5, 36, 38, 38, 36, 36, 42, 38, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 32, 32, 38, 32, 32, 38, 38, 36, 38, 42, 42, 42, 36, 36, 38, 38, 38, 38, 36, 36, 38, 38, 36, 36, 36, 38, 38, 38, 36, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 38, 32, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 36, 36, 32, 32, 32, 36, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 42, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 38, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 38, 43, 43, 43, 43, 43, 38, 38, 36, 32, 42, 42, 42, 36, 36, 36, 36, 36, 38, 36, 36, 42, 38, 42, 42, 44, 38, 38, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 36, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 7, 38, 38, 38, 38, 38, 38, 38, 43, 36, 36, 36, 36, 36, 36, 38, 36, 42, 42, 38, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 38, 38, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 38, 43, 43, 43, 43, 43, 38, 38, 36, 32, 45, 36, 42, 36, 36, 36, 36, 38, 38, 42, 42, 38, 38, 42, 42, 44, 38, 38, 38, 38, 38, 38, 38, 36, 36, 45, 38, 38, 38, 38, 43, 43, 38, 43, 32, 32, 36, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 19, 43, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 32, 38, 32, 32, 32, 32, 32, 32, 38, 38, 38, 32, 32, 32, 38, 32, 32, 32, 32, 38, 38, 38, 32, 32, 38, 32, 38, 32, 32, 38, 38, 38, 32, 32, 38, 38, 38, 32, 32, 32, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 45, 42, 36, 42, 42, 38, 38, 38, 42, 42, 42, 38, 42, 42, 42, 36, 38, 38, 32, 38, 38, 38, 38, 38, 38, 45, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 46, 46, 19, 19, 19, 19, 19, 19, 7, 19, 38, 38, 38, 38, 38, 36, 42, 42, 42, 36, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 38, 32, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 38, 36, 32, 36, 36, 36, 42, 42, 42, 42, 38, 36, 36, 36, 38, 36, 36, 36, 44, 38, 38, 38, 38, 38, 38, 38, 36, 36, 38, 43, 43, 43, 38, 32, 32, 38, 38, 32, 32, 36, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 5, 46, 46, 46, 46, 46, 46, 46, 19, 32, 36, 42, 42, 5, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 38, 38, 36, 32, 42, 36, 45, 42, 45, 42, 42, 38, 36, 45, 45, 38, 45, 45, 36, 36, 38, 38, 38, 38, 38, 38, 38, 45, 45, 38, 38, 38, 38, 38, 32, 32, 32, 38, 32, 32, 36, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 32, 32, 42, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 42, 42, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 38, 32, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 36, 32, 45, 42, 42, 36, 36, 36, 36, 38, 42, 42, 42, 38, 42, 42, 42, 44, 47, 19, 38, 38, 38, 38, 32, 32, 32, 45, 46, 46, 46, 46, 46, 46, 46, 32, 32, 32, 36, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 46, 46, 46, 46, 46, 46, 46, 46, 19, 32, 32, 32, 32, 32, 32, 38, 36, 42, 42, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 36, 38, 38, 38, 38, 45, 42, 42, 36, 36, 36, 38, 36, 38, 42, 42, 42, 42, 42, 42, 42, 45, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 42, 42, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 32, 48, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 7, 32, 32, 32, 32, 32, 32, 34, 36, 36, 36, 36, 36, 36, 36, 36, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 38, 32, 38, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 32, 48, 36, 36, 36, 36, 36, 36, 36, 36, 36, 32, 38, 38, 32, 32, 32, 32, 32, 38, 34, 38, 36, 36, 36, 36, 36, 36, 36, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 19, 19, 19, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 19, 5, 19, 19, 19, 36, 36, 19, 19, 19, 19, 19, 19, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 19, 36, 19, 36, 19, 36, 8, 9, 8, 9, 42, 42, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 36, 36, 36, 36, 36, 5, 36, 36, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 19, 19, 19, 19, 19, 19, 19, 19, 36, 19, 19, 19, 19, 19, 19, 38, 19, 19, 5, 5, 5, 5, 5, 19, 19, 19, 19, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 49, 49, 36, 36, 36, 36, 42, 36, 36, 36, 36, 36, 36, 49, 44, 36, 42, 42, 36, 36, 43, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 5, 5, 5, 5, 5, 43, 43, 43, 43, 43, 43, 42, 42, 36, 36, 43, 43, 43, 43, 36, 36, 36, 43, 49, 49, 49, 43, 43, 49, 49, 49, 49, 49, 49, 49, 43, 43, 43, 36, 36, 36, 36, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 49, 42, 36, 36, 49, 49, 49, 49, 49, 49, 36, 43, 49, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 49, 49, 49, 36, 19, 19, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 13, 38, 38, 38, 38, 38, 13, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 5, 34, 16, 16, 16, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 51, 52, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 38, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 38, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 36, 36, 36, 5, 5, 5, 5, 5, 5, 5, 5, 5, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 16, 16, 16, 16, 16, 16, 38, 38, 11, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 19, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 4, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 8, 9, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 5, 5, 5, 55, 55, 55, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 45, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 45, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 38, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 37, 37, 42, 36, 36, 36, 36, 36, 36, 36, 42, 42, 42, 42, 42, 42, 42, 42, 36, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 36, 44, 36, 5, 5, 5, 34, 5, 5, 5, 7, 32, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 5, 5, 11, 5, 5, 5, 5, 36, 36, 36, 24, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 36, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 32, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 36, 36, 36, 42, 42, 42, 42, 36, 36, 42, 42, 42, 38, 38, 38, 38, 42, 42, 36, 42, 42, 42, 42, 42, 42, 36, 36, 36, 38, 38, 38, 38, 19, 38, 38, 38, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 42, 42, 36, 38, 38, 5, 5, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 42, 36, 42, 36, 36, 36, 36, 36, 36, 36, 38, 44, 49, 36, 49, 49, 36, 36, 36, 36, 36, 36, 36, 36, 42, 42, 42, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 5, 5, 5, 34, 5, 5, 5, 5, 5, 5, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 39, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 42, 32, 32, 32, 32, 32, 32, 43, 43, 32, 32, 32, 32, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 45, 36, 36, 36, 36, 36, 45, 36, 45, 42, 42, 42, 42, 36, 45, 56, 43, 43, 43, 43, 43, 43, 43, 43, 38, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 5, 5, 5, 5, 5, 5, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 36, 36, 36, 36, 36, 36, 36, 36, 36, 19, 19, 19, 19, 19, 19, 19, 19, 19, 5, 5, 5, 36, 36, 42, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 42, 36, 36, 36, 36, 42, 42, 36, 36, 45, 44, 36, 36, 43, 43, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 43, 43, 43, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 42, 36, 36, 42, 42, 42, 36, 42, 36, 36, 36, 45, 45, 38, 38, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 42, 42, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 42, 42, 36, 36, 38, 38, 38, 5, 5, 5, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 32, 32, 32, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 34, 34, 34, 34, 34, 5, 5, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 16, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 13, 13, 13, 5, 5, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 5, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 36, 36, 36, 36, 36, 36, 36, 32, 32, 32, 32, 36, 32, 32, 32, 32, 32, 32, 36, 32, 32, 42, 36, 36, 32, 38, 38, 38, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 34, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 38, 38, 13, 13, 13, 13, 13, 13, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 38, 38, 13, 13, 13, 13, 13, 13, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 38, 13, 38, 13, 38, 13, 38, 13, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 33, 33, 33, 33, 33, 33, 33, 33, 16, 16, 16, 16, 16, 16, 16, 16, 33, 33, 33, 33, 33, 33, 33, 33, 16, 16, 16, 16, 16, 16, 16, 16, 33, 33, 33, 33, 33, 33, 33, 33, 16, 16, 16, 16, 16, 38, 16, 16, 13, 13, 13, 13, 33, 14, 16, 14, 14, 14, 16, 16, 16, 38, 16, 16, 13, 13, 13, 13, 33, 14, 14, 14, 16, 16, 16, 16, 38, 38, 16, 16, 13, 13, 13, 13, 38, 14, 14, 14, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 14, 14, 14, 38, 38, 16, 16, 16, 38, 16, 16, 13, 13, 13, 13, 33, 14, 14, 38, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 24, 57, 58, 24, 24, 59, 11, 11, 59, 59, 59, 17, 5, 60, 61, 8, 23, 60, 61, 8, 23, 17, 17, 17, 5, 17, 17, 17, 17, 62, 63, 24, 24, 24, 24, 24, 4, 17, 5, 17, 17, 5, 17, 5, 5, 5, 23, 29, 17, 64, 5, 17, 15, 15, 5, 5, 5, 10, 8, 9, 5, 5, 64, 5, 5, 5, 5, 5, 5, 5, 5, 10, 5, 15, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 24, 24, 24, 24, 24, 65, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 46, 34, 38, 38, 28, 46, 46, 46, 46, 46, 10, 10, 10, 8, 9, 35, 46, 28, 28, 28, 28, 46, 46, 46, 46, 46, 10, 10, 10, 8, 9, 38, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 38, 38, 38, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 18, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 39, 39, 39, 39, 36, 39, 39, 39, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 13, 26, 19, 26, 19, 13, 19, 26, 16, 13, 13, 13, 16, 16, 13, 13, 13, 31, 19, 13, 26, 19, 10, 13, 13, 13, 13, 13, 19, 19, 19, 26, 25, 19, 13, 19, 30, 19, 13, 19, 13, 30, 13, 13, 19, 16, 13, 13, 13, 13, 16, 32, 32, 32, 32, 66, 19, 19, 16, 16, 13, 13, 10, 10, 10, 10, 10, 13, 16, 16, 16, 16, 19, 10, 19, 19, 16, 19, 46, 46, 46, 28, 28, 46, 46, 46, 46, 46, 46, 28, 28, 28, 28, 46, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 55, 55, 55, 55, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 55, 55, 55, 55, 55, 55, 55, 55, 55, 13, 16, 55, 55, 55, 55, 28, 19, 19, 38, 38, 38, 38, 27, 27, 27, 27, 68, 25, 25, 25, 25, 25, 10, 10, 19, 19, 19, 19, 10, 19, 19, 10, 19, 19, 10, 19, 19, 21, 21, 19, 19, 19, 10, 19, 19, 19, 19, 19, 19, 19, 19, 19, 26, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 19, 19, 27, 19, 27, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 10, 27, 27, 10, 10, 10, 27, 27, 10, 10, 27, 10, 10, 10, 27, 10, 27, 10, 10, 10, 27, 10, 10, 10, 10, 27, 10, 10, 27, 27, 27, 27, 10, 10, 27, 10, 27, 10, 27, 27, 27, 27, 27, 27, 10, 27, 10, 10, 10, 10, 10, 27, 27, 27, 27, 10, 10, 10, 10, 27, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 10, 10, 10, 27, 10, 10, 10, 10, 10, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 27, 10, 10, 27, 27, 27, 27, 10, 10, 27, 27, 10, 10, 27, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 27, 10, 10, 27, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 10, 10, 10, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 27, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 19, 19, 19, 19, 19, 19, 19, 19, 8, 9, 8, 9, 19, 19, 19, 19, 19, 19, 26, 19, 19, 19, 19, 19, 19, 19, 69, 69, 19, 19, 19, 19, 10, 10, 19, 19, 19, 19, 19, 19, 21, 70, 71, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 10, 10, 10, 10, 19, 19, 19, 19, 19, 19, 19, 69, 69, 69, 69, 21, 21, 21, 69, 21, 21, 69, 19, 19, 19, 19, 21, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 25, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 46, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 19, 19, 19, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 19, 26, 26, 26, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 26, 26, 19, 26, 26, 26, 26, 26, 26, 26, 21, 21, 19, 19, 19, 19, 19, 19, 26, 26, 19, 19, 25, 27, 19, 19, 19, 19, 26, 26, 19, 19, 25, 27, 19, 19, 19, 19, 26, 26, 26, 19, 19, 26, 19, 19, 26, 26, 26, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 26, 26, 26, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 26, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 10, 72, 72, 73, 73, 10, 21, 21, 21, 21, 21, 26, 26, 19, 19, 26, 19, 19, 19, 19, 25, 26, 19, 21, 19, 19, 69, 69, 19, 19, 21, 19, 19, 19, 26, 69, 26, 19, 21, 19, 21, 21, 19, 19, 21, 19, 19, 19, 21, 19, 19, 19, 21, 21, 74, 74, 74, 74, 74, 74, 74, 74, 21, 21, 21, 19, 19, 19, 19, 19, 25, 19, 25, 19, 19, 19, 19, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 25, 26, 19, 25, 26, 25, 21, 26, 25, 26, 26, 19, 26, 26, 19, 27, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 19, 19, 21, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 74, 74, 74, 74, 74, 74, 19, 19, 21, 69, 21, 21, 21, 21, 19, 21, 19, 21, 21, 19, 26, 26, 21, 69, 19, 19, 19, 19, 19, 21, 19, 19, 69, 69, 19, 19, 19, 19, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 69, 26, 19, 19, 19, 19, 69, 69, 26, 26, 25, 26, 26, 26, 26, 26, 69, 25, 26, 25, 26, 25, 69, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 26, 19, 19, 19, 19, 26, 25, 69, 26, 26, 26, 26, 26, 25, 25, 69, 69, 25, 69, 26, 25, 25, 69, 69, 26, 26, 69, 26, 26, 19, 19, 21, 19, 19, 69, 19, 19, 21, 21, 69, 69, 69, 69, 19, 21, 19, 19, 21, 19, 21, 19, 21, 19, 19, 19, 19, 19, 19, 21, 19, 19, 19, 21, 19, 19, 19, 19, 19, 19, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 26, 19, 19, 19, 19, 19, 19, 21, 19, 19, 21, 19, 19, 19, 19, 69, 19, 69, 19, 19, 19, 19, 69, 69, 69, 19, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 21, 19, 19, 19, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 19, 69, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 10, 10, 10, 10, 10, 8, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 72, 72, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 8, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 8, 9, 8, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 8, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 19, 19, 19, 19, 19, 21, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 19, 19, 10, 10, 10, 10, 10, 10, 19, 19, 19, 69, 19, 19, 19, 19, 69, 26, 26, 26, 26, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 16, 13, 13, 13, 16, 16, 13, 16, 13, 16, 13, 16, 13, 13, 13, 13, 16, 13, 16, 16, 13, 16, 16, 16, 16, 16, 16, 34, 34, 13, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 19, 19, 19, 19, 19, 19, 13, 16, 13, 16, 36, 36, 36, 13, 16, 38, 38, 38, 38, 38, 5, 5, 5, 5, 46, 5, 5, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 16, 38, 38, 38, 38, 38, 16, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 34, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 5, 5, 23, 29, 23, 29, 5, 5, 5, 23, 29, 5, 23, 29, 5, 5, 5, 5, 5, 5, 5, 5, 5, 11, 5, 5, 11, 5, 23, 29, 5, 5, 23, 29, 8, 9, 8, 9, 8, 9, 8, 9, 5, 5, 5, 5, 5, 34, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 11, 11, 5, 5, 5, 5, 11, 5, 8, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 19, 19, 5, 5, 5, 8, 9, 8, 9, 8, 9, 8, 9, 11, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 75, 76, 76, 76, 74, 77, 78, 79, 70, 71, 70, 71, 70, 71, 70, 71, 70, 71, 74, 74, 70, 71, 70, 71, 70, 71, 70, 71, 80, 70, 71, 71, 74, 79, 79, 79, 79, 79, 79, 79, 79, 79, 81, 81, 81, 81, 82, 82, 83, 77, 77, 77, 77, 77, 74, 74, 79, 79, 79, 77, 78, 84, 74, 19, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 81, 81, 85, 85, 77, 77, 78, 80, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 76, 77, 77, 77, 78, 38, 38, 38, 38, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 86, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 74, 74, 87, 87, 87, 87, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 38, 38, 38, 38, 38, 38, 38, 38, 74, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 28, 28, 28, 28, 28, 28, 28, 28, 74, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 69, 74, 69, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 77, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 38, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 34, 34, 34, 34, 34, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 5, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 32, 36, 39, 39, 39, 5, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 5, 34, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 34, 34, 36, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 36, 36, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 34, 34, 34, 34, 34, 34, 34, 34, 34, 14, 14, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 16, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 34, 16, 16, 16, 16, 16, 16, 16, 16, 13, 16, 13, 16, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 34, 14, 14, 13, 16, 13, 16, 32, 13, 16, 13, 16, 16, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 13, 13, 13, 13, 16, 13, 13, 13, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 13, 13, 13, 16, 13, 16, 13, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 16, 13, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 34, 34, 34, 34, 13, 16, 32, 34, 34, 16, 32, 32, 32, 32, 32, 32, 32, 36, 32, 32, 32, 36, 32, 32, 32, 32, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 36, 36, 42, 19, 19, 19, 19, 36, 38, 38, 38, 46, 46, 46, 46, 46, 46, 19, 19, 7, 19, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 42, 42, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 32, 32, 32, 32, 32, 32, 5, 5, 5, 32, 5, 32, 32, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 45, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 5, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 38, 38, 38, 36, 36, 36, 42, 32, 32, 32, 32, 32, 43, 43, 43, 32, 32, 32, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 42, 42, 36, 36, 36, 36, 42, 42, 36, 36, 42, 42, 56, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 38, 34, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 5, 5, 43, 43, 43, 43, 43, 36, 34, 43, 43, 43, 43, 43, 43, 43, 43, 43, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 43, 43, 43, 43, 43, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 42, 42, 36, 36, 42, 42, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 36, 32, 32, 32, 32, 32, 32, 32, 32, 36, 42, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 5, 5, 5, 5, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 34, 43, 43, 43, 32, 32, 32, 19, 19, 19, 43, 49, 36, 49, 43, 43, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 32, 36, 36, 36, 32, 32, 36, 36, 32, 32, 32, 32, 32, 36, 36, 32, 36, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 34, 5, 5, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 42, 36, 36, 42, 42, 5, 5, 32, 34, 34, 42, 44, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 14, 34, 34, 34, 34, 16, 16, 16, 16, 16, 16, 16, 16, 16, 34, 14, 14, 38, 38, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 36, 42, 42, 36, 42, 42, 5, 42, 36, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 89, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 53, 38, 38, 38, 38, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 54, 38, 38, 38, 38, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 16, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 32, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 10, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 38, 32, 38, 32, 32, 38, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 9, 8, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 19, 19, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 19, 19, 19, 19, 19, 19, 19, 19, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 7, 19, 19, 19, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 37, 76, 76, 76, 76, 76, 76, 76, 70, 71, 76, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 76, 80, 80, 94, 94, 70, 71, 70, 71, 70, 71, 70, 71, 70, 71, 70, 71, 70, 71, 70, 71, 76, 76, 70, 71, 76, 76, 76, 76, 94, 94, 94, 76, 76, 76, 38, 76, 76, 76, 76, 80, 70, 71, 70, 71, 70, 71, 76, 76, 76, 95, 80, 95, 95, 95, 38, 76, 96, 76, 76, 38, 38, 38, 38, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 24, 38, 76, 76, 76, 96, 76, 76, 76, 70, 71, 76, 95, 76, 80, 76, 76, 97, 97, 97, 97, 97, 97, 97, 97, 97, 97, 76, 76, 95, 95, 95, 76, 76, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 70, 76, 71, 85, 94, 85, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 70, 95, 71, 95, 70, 71, 5, 8, 9, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 100, 100, 101, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 38, 38, 38, 96, 96, 95, 85, 74, 96, 96, 38, 19, 10, 10, 10, 10, 19, 19, 38, 65, 65, 65, 65, 65, 65, 65, 65, 65, 24, 24, 24, 19, 26, 93, 93, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 5, 5, 5, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 46, 46, 46, 46, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 46, 46, 19, 19, 19, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 55, 32, 32, 32, 32, 32, 32, 32, 32, 55, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 5, 55, 55, 55, 55, 55, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 5, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 13, 13, 13, 13, 13, 13, 13, 38, 13, 13, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 16, 16, 16, 16, 16, 16, 16, 38, 16, 16, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 34, 34, 34, 34, 34, 34, 38, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 38, 34, 34, 34, 34, 34, 34, 34, 34, 34, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 38, 38, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 38, 38, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 5, 46, 46, 46, 46, 46, 46, 46, 46, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 19, 19, 46, 46, 46, 46, 46, 46, 46, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 46, 46, 46, 46, 46, 46, 38, 38, 38, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 46, 46, 32, 32, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 43, 36, 36, 36, 38, 36, 36, 38, 38, 38, 38, 38, 36, 36, 36, 36, 43, 43, 43, 43, 38, 43, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 38, 36, 36, 36, 38, 38, 38, 38, 44, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 46, 46, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 19, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 38, 38, 38, 38, 46, 46, 46, 46, 46, 5, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 5, 5, 5, 5, 5, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 32, 32, 32, 34, 32, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 38, 36, 36, 36, 36, 36, 11, 34, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 38, 38, 38, 10, 10, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 36, 36, 11, 38, 38, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 34, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 5, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 32, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 46, 46, 46, 46, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 42, 36, 42, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 5, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 36, 32, 32, 36, 36, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 42, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 36, 42, 42, 36, 36, 5, 5, 40, 5, 5, 5, 5, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 40, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 36, 36, 36, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 36, 36, 36, 36, 42, 36, 36, 36, 36, 36, 36, 44, 36, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 5, 5, 5, 43, 42, 42, 43, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 5, 5, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 42, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 45, 32, 47, 47, 32, 5, 5, 5, 5, 36, 36, 36, 36, 5, 42, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 32, 5, 32, 5, 5, 5, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 42, 42, 36, 45, 36, 36, 5, 5, 5, 5, 5, 5, 36, 32, 32, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 38, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 5, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 36, 36, 42, 42, 38, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 32, 32, 32, 32, 32, 38, 36, 36, 32, 45, 42, 36, 42, 42, 42, 42, 38, 38, 42, 42, 38, 38, 42, 42, 45, 38, 38, 32, 38, 38, 38, 38, 38, 38, 45, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 42, 42, 38, 38, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 38, 38, 43, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 32, 45, 42, 42, 36, 36, 36, 36, 36, 36, 38, 45, 38, 38, 45, 38, 45, 45, 45, 42, 38, 42, 42, 36, 45, 44, 47, 36, 32, 5, 5, 38, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 42, 42, 36, 36, 36, 42, 36, 32, 32, 32, 32, 5, 5, 5, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 5, 5, 38, 5, 36, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 45, 42, 42, 36, 36, 36, 36, 36, 36, 42, 36, 42, 42, 45, 42, 36, 36, 42, 36, 36, 32, 32, 5, 32, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 45, 42, 42, 36, 36, 36, 36, 38, 38, 42, 42, 42, 42, 36, 36, 42, 36, 36, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 32, 32, 32, 32, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 42, 42, 36, 42, 36, 36, 5, 5, 5, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 42, 36, 42, 42, 36, 36, 36, 36, 36, 36, 45, 36, 32, 5, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 36, 42, 36, 49, 49, 36, 36, 36, 36, 42, 36, 36, 36, 36, 36, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 46, 5, 5, 5, 19, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 36, 36, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 43, 43, 43, 43, 43, 43, 43, 38, 38, 43, 38, 38, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 45, 42, 42, 42, 42, 42, 38, 42, 42, 38, 38, 36, 36, 45, 44, 47, 42, 47, 42, 36, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 36, 36, 36, 36, 38, 38, 36, 36, 42, 42, 42, 42, 36, 32, 5, 32, 42, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 43, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 36, 36, 36, 36, 36, 36, 42, 32, 36, 36, 36, 36, 5, 5, 5, 5, 5, 5, 5, 5, 44, 38, 38, 38, 38, 38, 38, 38, 38, 43, 36, 36, 36, 36, 36, 36, 42, 42, 36, 36, 36, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 47, 47, 47, 47, 47, 47, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 36, 44, 5, 5, 5, 32, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 42, 36, 36, 36, 42, 36, 42, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 36, 36, 36, 36, 36, 36, 36, 38, 36, 36, 36, 36, 36, 36, 42, 36, 32, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 42, 36, 36, 36, 36, 36, 36, 36, 42, 36, 36, 42, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 38, 38, 38, 36, 38, 36, 36, 38, 36, 36, 36, 36, 36, 36, 36, 47, 36, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 42, 42, 42, 42, 42, 38, 36, 36, 38, 42, 42, 36, 42, 36, 32, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 32, 32, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 42, 42, 5, 5, 38, 38, 38, 38, 38, 38, 38, 36, 36, 47, 42, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 38, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 42, 42, 36, 36, 36, 36, 36, 38, 38, 38, 42, 42, 36, 45, 44, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 19, 19, 19, 19, 19, 19, 19, 19, 7, 7, 7, 7, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 38, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 36, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 42, 42, 42, 36, 36, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 5, 5, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 36, 36, 36, 36, 36, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 36, 36, 36, 5, 5, 5, 5, 5, 19, 19, 19, 19, 34, 34, 34, 34, 5, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 46, 46, 46, 46, 46, 46, 46, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 34, 34, 34, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 53, 32, 32, 32, 53, 53, 53, 53, 34, 34, 5, 5, 5, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 5, 5, 5, 5, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 36, 32, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 42, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 77, 77, 76, 77, 81, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 82, 82, 77, 77, 79, 79, 79, 38, 38, 38, 38, 38, 38, 38, 38, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 77, 77, 77, 77, 38, 77, 77, 77, 77, 77, 77, 77, 38, 77, 77, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 78, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 78, 78, 78, 38, 38, 78, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 78, 78, 78, 78, 38, 38, 38, 38, 38, 38, 38, 38, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 19, 36, 36, 5, 24, 24, 24, 24, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 19, 19, 19, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 10, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 45, 45, 36, 36, 36, 19, 19, 19, 45, 45, 45, 45, 45, 45, 24, 24, 24, 24, 24, 24, 24, 24, 36, 36, 36, 36, 36, 36, 36, 36, 19, 19, 36, 36, 36, 36, 36, 36, 36, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 36, 36, 36, 36, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 36, 36, 36, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 38, 38, 38, 38, 38, 38, 38, 38, 38, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 87, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 38, 13, 13, 38, 38, 13, 38, 38, 13, 13, 38, 38, 13, 13, 13, 13, 38, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 38, 16, 38, 16, 16, 16, 16, 16, 16, 16, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 38, 13, 13, 13, 13, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 38, 13, 13, 13, 13, 13, 13, 13, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 38, 13, 13, 13, 13, 38, 13, 13, 13, 13, 13, 38, 13, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 10, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 10, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 10, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 10, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 10, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 10, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 10, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 10, 16, 16, 16, 16, 16, 16, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 10, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 10, 16, 16, 16, 16, 16, 16, 13, 16, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 19, 19, 19, 19, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 19, 19, 19, 19, 19, 19, 19, 19, 36, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 36, 19, 19, 5, 5, 5, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 32, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 38, 16, 16, 16, 16, 16, 16, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 38, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 38, 38, 36, 36, 36, 36, 36, 36, 36, 38, 36, 36, 38, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 36, 36, 36, 36, 36, 36, 36, 34, 34, 34, 34, 34, 34, 34, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 32, 19, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 36, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 7, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 34, 36, 36, 36, 36, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 36, 36, 32, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 36, 32, 32, 36, 32, 32, 32, 32, 32, 32, 32, 36, 36, 32, 32, 32, 32, 32, 36, 38, 38, 38, 38, 38, 38, 38, 38, 32, 34, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 36, 36, 36, 36, 36, 36, 36, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 36, 36, 36, 36, 36, 36, 36, 34, 38, 38, 38, 38, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 38, 38, 38, 38, 5, 5, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 19, 46, 46, 46, 7, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 19, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 38, 32, 38, 38, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 32, 38, 32, 38, 38, 38, 38, 38, 38, 32, 38, 38, 38, 38, 32, 38, 32, 38, 32, 38, 32, 32, 32, 38, 32, 32, 38, 32, 38, 38, 32, 38, 32, 38, 32, 38, 32, 38, 32, 38, 32, 32, 38, 32, 38, 38, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 32, 32, 32, 32, 38, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 32, 32, 32, 38, 32, 32, 32, 32, 32, 38, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 10, 10, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 19, 19, 19, 19, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 46, 46, 19, 19, 19, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 19, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 19, 19, 19, 19, 19, 25, 25, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 25, 25, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 69, 26, 26, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 19, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 74, 69, 69, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 69, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 69, 74, 74, 69, 69, 69, 69, 69, 69, 69, 69, 69, 74, 102, 102, 102, 102, 74, 74, 74, 74, 74, 74, 74, 74, 74, 102, 102, 102, 102, 102, 102, 102, 69, 69, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 74, 74, 74, 74, 74, 74, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 21, 19, 19, 21, 21, 21, 21, 21, 21, 21, 21, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 21, 21, 19, 21, 21, 21, 19, 19, 21, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 21, 21, 69, 69, 69, 69, 69, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 21, 69, 21, 19, 21, 69, 69, 69, 104, 104, 104, 104, 104, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 21, 69, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 21, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 21, 69, 69, 69, 69, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 19, 19, 19, 19, 19, 21, 21, 19, 19, 21, 69, 69, 21, 21, 21, 21, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 19, 19, 21, 21, 21, 21, 19, 19, 69, 19, 19, 19, 19, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 21, 19, 19, 21, 19, 19, 19, 19, 19, 19, 19, 19, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 19, 19, 19, 19, 19, 21, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 21, 21, 21, 19, 19, 19, 19, 19, 19, 19, 19, 21, 21, 21, 19, 19, 21, 19, 21, 19, 19, 19, 19, 21, 19, 19, 19, 19, 19, 19, 21, 19, 19, 19, 21, 19, 19, 19, 19, 19, 19, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 19, 19, 19, 21, 69, 21, 21, 21, 69, 69, 69, 19, 19, 69, 69, 69, 69, 102, 102, 102, 69, 69, 69, 69, 21, 21, 21, 21, 21, 21, 19, 19, 19, 21, 19, 69, 69, 102, 102, 102, 21, 19, 19, 21, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 102, 102, 69, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 19, 19, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 10, 10, 10, 10, 10, 10, 10, 10, 10, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 69, 102, 102, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 102, 102, 69, 69, 69, 69, 69, 69, 69, 69, 69, 69, 102, 102, 102, 102, 102, 102, 102, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 38, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 19, 38, 38, 38, 38, 38, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 102, 93, 93, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 93, 93, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 78, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 38, 93, 93, 65, 24, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 105, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 93, 93
};
static const CharProps CharProps_t3[106] = {
	{.shifted_width=4, .is_emoji=0, .category=UC_Cc, .is_emoji_presentation_base=0, .is_invalid=1, .is_non_rendered=1, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Control, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 0
	{.shifted_width=3, .is_emoji=0, .category=UC_Cc, .is_emoji_presentation_base=0, .is_invalid=1, .is_non_rendered=1, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Control, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 1
	{.shifted_width=3, .is_emoji=0, .category=UC_Cc, .is_emoji_presentation_base=0, .is_invalid=1, .is_non_rendered=1, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_LF, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 2
	{.shifted_width=3, .is_emoji=0, .category=UC_Cc, .is_emoji_presentation_base=0, .is_invalid=1, .is_non_rendered=1, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_CR, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 3
	{.shifted_width=5, .is_emoji=0, .category=UC_Zs, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 4
	{.shifted_width=5, .is_emoji=0, .category=UC_Po, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 5
	{.shifted_width=5, .is_emoji=1, .category=UC_Po, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 6
	{.shifted_width=5, .is_emoji=0, .category=UC_Sc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 7
	{.shifted_width=5, .is_emoji=0, .category=UC_Ps, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 8
	{.shifted_width=5, .is_emoji=0, .category=UC_Pe, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 9
	{.shifted_width=5, .is_emoji=0, .category=UC_Sm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 10
	{.shifted_width=5, .is_emoji=0, .category=UC_Pd, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 11
	{.shifted_width=5, .is_emoji=1, .category=UC_Nd, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 12
	{.shifted_width=5, .is_emoji=0, .category=UC_Lu, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 13
	{.shifted_width=5, .is_emoji=0, .category=UC_Sk, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 14
	{.shifted_width=5, .is_emoji=0, .category=UC_Pc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 15
	{.shifted_width=5, .is_emoji=0, .category=UC_Ll, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 16
	{.shifted_width=2, .is_emoji=0, .category=UC_Po, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 17
	{.shifted_width=2, .is_emoji=0, .category=UC_Sc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 18
	{.shifted_width=5, .is_emoji=0, .category=UC_So, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 19
	{.shifted_width=2, .is_emoji=0, .category=UC_Sk, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 20
	{.shifted_width=5, .is_emoji=1, .category=UC_So, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 21
	{.shifted_width=2, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 22
	{.shifted_width=5, .is_emoji=0, .category=UC_Pi, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 23
	{.shifted_width=4, .is_emoji=0, .category=UC_Cf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Control, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 24
	{.shifted_width=2, .is_emoji=1, .category=UC_So, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 25
	{.shifted_width=2, .is_emoji=0, .category=UC_So, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 26
	{.shifted_width=2, .is_emoji=0, .category=UC_Sm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 27
	{.shifted_width=2, .is_emoji=0, .category=UC_No, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 28
	{.shifted_width=5, .is_emoji=0, .category=UC_Pf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 29
	{.shifted_width=2, .is_emoji=0, .category=UC_Lu, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 30
	{.shifted_width=2, .is_emoji=0, .category=UC_Ll, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 31
	{.shifted_width=5, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 32
	{.shifted_width=5, .is_emoji=0, .category=UC_Lt, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 33
	{.shifted_width=5, .is_emoji=0, .category=UC_Lm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 34
	{.shifted_width=2, .is_emoji=0, .category=UC_Lm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 35
	{.shifted_width=4, .is_emoji=0, .category=UC_Mn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 36
	{.shifted_width=4, .is_emoji=0, .category=UC_Mn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 37
	{.shifted_width=0, .is_emoji=0, .category=UC_Cn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 38
	{.shifted_width=4, .is_emoji=0, .category=UC_Me, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 39
	{.shifted_width=4, .is_emoji=0, .category=UC_Cf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Prepend, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 40
	{.shifted_width=5, .is_emoji=0, .category=UC_Nd, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 41
	{.shifted_width=4, .is_emoji=0, .category=UC_Mc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_SpacingMark, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 42
	{.shifted_width=5, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_Consonant, .is_extended_pictographic=0}, // 43
	{.shifted_width=4, .is_emoji=0, .category=UC_Mn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Linker, .is_extended_pictographic=0}, // 44
	{.shifted_width=4, .is_emoji=0, .category=UC_Mc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 45
	{.shifted_width=5, .is_emoji=0, .category=UC_No, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 46
	{.shifted_width=5, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_Prepend, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 47
	{.shifted_width=5, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_SpacingMark, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 48
	{.shifted_width=4, .is_emoji=0, .category=UC_Mc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 49
	{.shifted_width=6, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_L, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 50
	{.shifted_width=6, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_L, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 51
	{.shifted_width=4, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_V, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 52
	{.shifted_width=5, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_V, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 53
	{.shifted_width=5, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_T, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 54
	{.shifted_width=5, .is_emoji=0, .category=UC_Nl, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 55
	{.shifted_width=4, .is_emoji=0, .category=UC_Mc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Linker, .is_extended_pictographic=0}, // 56
	{.shifted_width=4, .is_emoji=0, .category=UC_Cf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 57
	{.shifted_width=4, .is_emoji=0, .category=UC_Cf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_ZWJ, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 58
	{.shifted_width=2, .is_emoji=0, .category=UC_Pd, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 59
	{.shifted_width=2, .is_emoji=0, .category=UC_Pi, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 60
	{.shifted_width=2, .is_emoji=0, .category=UC_Pf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 61
	{.shifted_width=5, .is_emoji=0, .category=UC_Zl, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Control, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 62
	{.shifted_width=5, .is_emoji=0, .category=UC_Zp, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Control, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 63
	{.shifted_width=5, .is_emoji=1, .category=UC_Po, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 64
	{.shifted_width=4, .is_emoji=0, .category=UC_Cn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Control, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 65
	{.shifted_width=5, .is_emoji=1, .category=UC_Ll, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 66
	{.shifted_width=2, .is_emoji=0, .category=UC_Nl, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 67
	{.shifted_width=2, .is_emoji=1, .category=UC_Sm, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 68
	{.shifted_width=6, .is_emoji=1, .category=UC_So, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 69
	{.shifted_width=6, .is_emoji=0, .category=UC_Ps, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 70
	{.shifted_width=6, .is_emoji=0, .category=UC_Pe, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 71
	{.shifted_width=5, .is_emoji=1, .category=UC_Sm, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 72
	{.shifted_width=6, .is_emoji=1, .category=UC_Sm, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 73
	{.shifted_width=6, .is_emoji=0, .category=UC_So, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 74
	{.shifted_width=6, .is_emoji=0, .category=UC_Zs, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 75
	{.shifted_width=6, .is_emoji=0, .category=UC_Po, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 76
	{.shifted_width=6, .is_emoji=0, .category=UC_Lm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 77
	{.shifted_width=6, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 78
	{.shifted_width=6, .is_emoji=0, .category=UC_Nl, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 79
	{.shifted_width=6, .is_emoji=0, .category=UC_Pd, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 80
	{.shifted_width=6, .is_emoji=0, .category=UC_Mn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 81
	{.shifted_width=6, .is_emoji=0, .category=UC_Mc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 82
	{.shifted_width=6, .is_emoji=1, .category=UC_Pd, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 83
	{.shifted_width=6, .is_emoji=1, .category=UC_Po, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 84
	{.shifted_width=6, .is_emoji=0, .category=UC_Sk, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 85
	{.shifted_width=6, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 86
	{.shifted_width=6, .is_emoji=0, .category=UC_No, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 87
	{.shifted_width=6, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_LV, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 88
	{.shifted_width=6, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_LVT, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 89
	{.shifted_width=3, .is_emoji=0, .category=UC_Cs, .is_emoji_presentation_base=0, .is_invalid=1, .is_non_rendered=1, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 90
	{.shifted_width=2, .is_emoji=0, .category=UC_Co, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 91
	{.shifted_width=6, .is_emoji=0, .category=UC_Cn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 92
	{.shifted_width=3, .is_emoji=0, .category=UC_Cn, .is_emoji_presentation_base=0, .is_invalid=1, .is_non_rendered=1, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 93
	{.shifted_width=6, .is_emoji=0, .category=UC_Pc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=1, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 94
	{.shifted_width=6, .is_emoji=0, .category=UC_Sm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 95
	{.shifted_width=6, .is_emoji=0, .category=UC_Sc, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 96
	{.shifted_width=6, .is_emoji=0, .category=UC_Nd, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 97
	{.shifted_width=6, .is_emoji=0, .category=UC_Lu, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 98
	{.shifted_width=6, .is_emoji=0, .category=UC_Ll, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 99
	{.shifted_width=5, .is_emoji=0, .category=UC_Lm, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 100
	{.shifted_width=4, .is_emoji=0, .category=UC_Lo, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=1, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 101
	{.shifted_width=0, .is_emoji=0, .category=UC_Cn, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=0, .is_symbol=0, .is_combining_char=0, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_None, .indic_conjunct_break=ICB_None, .is_extended_pictographic=1}, // 102
	{.shifted_width=6, .is_emoji=1, .category=UC_So, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Regional_Indicator, .indic_conjunct_break=ICB_None, .is_extended_pictographic=0}, // 103
	{.shifted_width=6, .is_emoji=1, .category=UC_Sk, .is_emoji_presentation_base=1, .is_invalid=0, .is_non_rendered=0, .is_symbol=1, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 104
	{.shifted_width=4, .is_emoji=0, .category=UC_Cf, .is_emoji_presentation_base=0, .is_invalid=0, .is_non_rendered=1, .is_symbol=0, .is_combining_char=1, .is_word_char=0, .is_punctuation=0, .grapheme_break=GBP_Extend, .indic_conjunct_break=ICB_Extend, .is_extended_pictographic=0}, // 105
};
static const char_type GraphemeSegmentationResult_mask = 15u;
static const char_type GraphemeSegmentationResult_shift = 4u;
static const uint8_t GraphemeSegmentationResult_t1[4096] = {
	0, 0, 1, 0, 2, 2, 3, 2, 4, 4, 5, 4, 6, 6, 7, 6, 8, 8, 9, 8, 10, 10, 11, 10, 12, 12, 13, 12, 14, 14, 15, 14, 16, 16, 17, 16, 18, 18, 19, 18, 16, 16, 17, 16, 18, 18, 19, 18, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 20, 20, 21, 20, 22, 22, 23, 22, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 4, 4, 5, 4, 6, 6, 7, 6, 32, 32, 33, 32, 34, 34, 35, 34, 0, 36, 1, 1, 2, 37, 3, 3, 4, 38, 5, 5, 6, 39, 7, 7, 8, 40, 9, 9, 10, 41, 11, 11, 12, 42, 13, 13, 14, 43, 15, 15, 16, 44, 17, 17, 18, 45, 19, 19, 16, 44, 17, 17, 18, 45, 19, 19, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 20, 46, 21, 21, 22, 47, 23, 23, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 4, 38, 5, 5, 6, 39, 7, 7, 32, 52, 33, 33, 34, 53, 35, 35, 0, 0, 1, 0, 2, 2, 3, 2, 4, 4, 5, 4, 6, 6, 7, 6, 8, 8, 9, 8, 10, 10, 11, 10, 12, 12, 13, 12, 14, 14, 15, 14, 16, 16, 17, 16, 18, 18, 19, 18, 16, 16, 17, 16, 18, 18, 19, 18, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 20, 20, 21, 20, 22, 22, 23, 22, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 4, 4, 5, 4, 6, 6, 7, 6, 32, 32, 33, 32, 34, 34, 35, 34, 0, 36, 1, 1, 2, 37, 3, 3, 4, 38, 5, 5, 6, 39, 7, 7, 8, 40, 9, 9, 10, 41, 11, 11, 12, 42, 13, 13, 14, 43, 15, 15, 16, 44, 17, 17, 18, 45, 19, 19, 16, 44, 17, 17, 18, 45, 19, 19, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 20, 46, 21, 21, 22, 47, 23, 23, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 4, 38, 5, 5, 6, 39, 7, 7, 32, 52, 33, 33, 34, 53, 35, 35, 0, 54, 1, 54, 2, 55, 3, 55, 4, 56, 9, 56, 6, 57, 11, 57, 8, 58, 9, 58, 10, 59, 11, 59, 12, 60, 13, 60, 14, 61, 15, 61, 16, 62, 17, 62, 18, 63, 19, 63, 16, 62, 17, 62, 18, 63, 19, 63, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 20, 64, 9, 64, 22, 65, 11, 65, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 4, 56, 9, 56, 6, 57, 11, 57, 32, 70, 9, 70, 34, 71, 11, 71, 0, 36, 1, 72, 2, 37, 3, 73, 4, 38, 9, 74, 6, 39, 11, 75, 8, 40, 9, 76, 10, 41, 11, 77, 12, 42, 13, 78, 14, 43, 15, 79, 16, 44, 17, 80, 18, 45, 19, 81, 16, 44, 17, 80, 18, 45, 19, 81, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 20, 46, 9, 82, 22, 47, 11, 83, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 4, 38, 9, 74, 6, 39, 11, 75, 32, 52, 9, 88, 34, 53, 11, 89, 0, 54, 1, 54, 2, 55, 3, 55, 4, 56, 9, 56, 6, 57, 11, 57, 8, 58, 9, 58, 10, 59, 11, 59, 12, 60, 13, 60, 14, 61, 15, 61, 16, 62, 17, 62, 18, 63, 19, 63, 16, 62, 17, 62, 18, 63, 19, 63, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 20, 64, 9, 64, 22, 65, 11, 65, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 4, 56, 9, 56, 6, 57, 11, 57, 32, 70, 9, 70, 34, 71, 11, 71, 0, 36, 1, 72, 2, 37, 3, 73, 4, 38, 9, 74, 6, 39, 11, 75, 8, 40, 9, 76, 10, 41, 11, 77, 12, 42, 13, 78, 14, 43, 15, 79, 16, 44, 17, 80, 18, 45, 19, 81, 16, 44, 17, 80, 18, 45, 19, 81, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 20, 46, 9, 82, 22, 47, 11, 83, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 4, 38, 9, 74, 6, 39, 11, 75, 32, 52, 9, 88, 34, 53, 11, 89, 90, 90, 91, 90, 92, 92, 93, 92, 94, 94, 95, 94, 96, 96, 97, 96, 98, 98, 99, 98, 100, 100, 101, 100, 102, 102, 103, 102, 104, 104, 105, 104, 106, 106, 107, 106, 108, 108, 109, 108, 106, 106, 107, 106, 108, 108, 109, 108, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 110, 110, 111, 110, 112, 112, 113, 112, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 94, 94, 95, 94, 96, 96, 97, 96, 122, 122, 123, 122, 124, 124, 125, 124, 90, 126, 91, 91, 92, 127, 93, 93, 94, 128, 95, 95, 96, 129, 97, 97, 98, 130, 99, 99, 100, 131, 101, 101, 102, 132, 103, 103, 104, 133, 105, 105, 106, 134, 107, 107, 108, 135, 109, 109, 106, 134, 107, 107, 108, 135, 109, 109, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 110, 136, 111, 111, 112, 137, 113, 113, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 94, 128, 95, 95, 96, 129, 97, 97, 122, 142, 123, 123, 124, 143, 125, 125, 90, 90, 91, 90, 92, 92, 93, 92, 94, 94, 95, 94, 96, 96, 97, 96, 98, 98, 99, 98, 100, 100, 101, 100, 102, 102, 103, 102, 104, 104, 105, 104, 106, 106, 107, 106, 108, 108, 109, 108, 106, 106, 107, 106, 108, 108, 109, 108, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 110, 110, 111, 110, 112, 112, 113, 112, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 94, 94, 95, 94, 96, 96, 97, 96, 122, 122, 123, 122, 124, 124, 125, 124, 90, 126, 91, 91, 92, 127, 93, 93, 94, 128, 95, 95, 96, 129, 97, 97, 98, 130, 99, 99, 100, 131, 101, 101, 102, 132, 103, 103, 104, 133, 105, 105, 106, 134, 107, 107, 108, 135, 109, 109, 106, 134, 107, 107, 108, 135, 109, 109, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 110, 136, 111, 111, 112, 137, 113, 113, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 94, 128, 95, 95, 96, 129, 97, 97, 122, 142, 123, 123, 124, 143, 125, 125, 90, 144, 91, 144, 92, 145, 93, 145, 94, 146, 99, 146, 96, 147, 101, 147, 98, 148, 99, 148, 100, 149, 101, 149, 102, 150, 103, 150, 104, 151, 105, 151, 106, 152, 107, 152, 108, 153, 109, 153, 106, 152, 107, 152, 108, 153, 109, 153, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 110, 154, 99, 154, 112, 155, 101, 155, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 94, 146, 99, 146, 96, 147, 101, 147, 122, 160, 99, 160, 124, 161, 101, 161, 90, 126, 91, 162, 92, 127, 93, 163, 94, 128, 99, 164, 96, 129, 101, 165, 98, 130, 99, 166, 100, 131, 101, 167, 102, 132, 103, 168, 104, 133, 105, 169, 106, 134, 107, 170, 108, 135, 109, 171, 106, 134, 107, 170, 108, 135, 109, 171, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 110, 136, 99, 172, 112, 137, 101, 173, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 94, 128, 99, 164, 96, 129, 101, 165, 122, 142, 99, 178, 124, 143, 101, 179, 90, 144, 91, 144, 92, 145, 93, 145, 94, 146, 99, 146, 96, 147, 101, 147, 98, 148, 99, 148, 100, 149, 101, 149, 102, 150, 103, 150, 104, 151, 105, 151, 106, 152, 107, 152, 108, 153, 109, 153, 106, 152, 107, 152, 108, 153, 109, 153, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 110, 154, 99, 154, 112, 155, 101, 155, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 94, 146, 99, 146, 96, 147, 101, 147, 122, 160, 99, 160, 124, 161, 101, 161, 90, 126, 91, 162, 92, 127, 93, 163, 94, 128, 99, 164, 96, 129, 101, 165, 98, 130, 99, 166, 100, 131, 101, 167, 102, 132, 103, 168, 104, 133, 105, 169, 106, 134, 107, 170, 108, 135, 109, 171, 106, 134, 107, 170, 108, 135, 109, 171, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 110, 136, 99, 172, 112, 137, 101, 173, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 94, 128, 99, 164, 96, 129, 101, 165, 122, 142, 99, 178, 124, 143, 101, 179, 0, 0, 1, 0, 2, 2, 3, 2, 4, 4, 5, 4, 6, 6, 7, 6, 8, 8, 9, 8, 10, 10, 11, 10, 12, 12, 13, 12, 14, 14, 15, 14, 16, 16, 17, 16, 18, 18, 19, 18, 16, 16, 17, 16, 18, 18, 19, 18, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 20, 20, 21, 20, 22, 22, 23, 22, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 4, 4, 5, 4, 10, 10, 11, 10, 32, 32, 33, 32, 34, 34, 35, 34, 0, 36, 1, 1, 2, 37, 3, 3, 4, 38, 5, 5, 6, 39, 7, 7, 8, 40, 9, 9, 10, 41, 11, 11, 12, 42, 13, 13, 14, 43, 15, 15, 16, 44, 17, 17, 18, 45, 19, 19, 16, 44, 17, 17, 18, 45, 19, 19, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 20, 46, 21, 21, 22, 47, 23, 23, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 4, 38, 5, 5, 10, 41, 11, 11, 32, 52, 33, 33, 34, 53, 35, 35, 0, 0, 1, 0, 2, 2, 3, 2, 4, 4, 5, 4, 6, 6, 7, 6, 8, 8, 9, 8, 10, 10, 11, 10, 12, 12, 13, 12, 14, 14, 15, 14, 16, 16, 17, 16, 18, 18, 19, 18, 16, 16, 17, 16, 18, 18, 19, 18, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 4, 4, 5, 4, 6, 6, 7, 6, 20, 20, 21, 20, 22, 22, 23, 22, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 24, 24, 25, 24, 26, 26, 27, 26, 28, 28, 29, 28, 30, 30, 31, 30, 4, 4, 5, 4, 10, 10, 11, 10, 32, 32, 33, 32, 34, 34, 35, 34, 0, 36, 1, 1, 2, 37, 3, 3, 4, 38, 5, 5, 6, 39, 7, 7, 8, 40, 9, 9, 10, 41, 11, 11, 12, 42, 13, 13, 14, 43, 15, 15, 16, 44, 17, 17, 18, 45, 19, 19, 16, 44, 17, 17, 18, 45, 19, 19, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 4, 38, 5, 5, 6, 39, 7, 7, 20, 46, 21, 21, 22, 47, 23, 23, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 24, 48, 25, 25, 26, 49, 27, 27, 28, 50, 29, 29, 30, 51, 31, 31, 4, 38, 5, 5, 10, 41, 11, 11, 32, 52, 33, 33, 34, 53, 35, 35, 0, 54, 1, 54, 2, 55, 3, 55, 4, 56, 9, 56, 6, 57, 11, 57, 8, 58, 9, 58, 10, 59, 11, 59, 12, 60, 13, 60, 14, 61, 15, 61, 16, 62, 17, 62, 18, 63, 19, 63, 16, 62, 17, 62, 18, 63, 19, 63, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 20, 64, 9, 64, 22, 65, 11, 65, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 4, 56, 9, 56, 10, 59, 11, 59, 32, 70, 9, 70, 34, 71, 11, 71, 0, 36, 1, 72, 2, 37, 3, 73, 4, 38, 9, 74, 6, 39, 11, 75, 8, 40, 9, 76, 10, 41, 11, 77, 12, 42, 13, 78, 14, 43, 15, 79, 16, 44, 17, 80, 18, 45, 19, 81, 16, 44, 17, 80, 18, 45, 19, 81, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 20, 46, 9, 82, 22, 47, 11, 83, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 4, 38, 9, 74, 10, 41, 11, 77, 32, 52, 9, 88, 34, 53, 11, 89, 0, 54, 1, 54, 2, 55, 3, 55, 4, 56, 9, 56, 6, 57, 11, 57, 8, 58, 9, 58, 10, 59, 11, 59, 12, 60, 13, 60, 14, 61, 15, 61, 16, 62, 17, 62, 18, 63, 19, 63, 16, 62, 17, 62, 18, 63, 19, 63, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 4, 56, 9, 56, 6, 57, 11, 57, 20, 64, 9, 64, 22, 65, 11, 65, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 24, 66, 9, 66, 26, 67, 11, 67, 28, 68, 9, 68, 30, 69, 11, 69, 4, 56, 9, 56, 10, 59, 11, 59, 32, 70, 9, 70, 34, 71, 11, 71, 0, 36, 1, 72, 2, 37, 3, 73, 4, 38, 9, 74, 6, 39, 11, 75, 8, 40, 9, 76, 10, 41, 11, 77, 12, 42, 13, 78, 14, 43, 15, 79, 16, 44, 17, 80, 18, 45, 19, 81, 16, 44, 17, 80, 18, 45, 19, 81, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 4, 38, 9, 74, 6, 39, 11, 75, 20, 46, 9, 82, 22, 47, 11, 83, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 24, 48, 9, 84, 26, 49, 11, 85, 28, 50, 9, 86, 30, 51, 11, 87, 4, 38, 9, 74, 10, 41, 11, 77, 32, 52, 9, 88, 34, 53, 11, 89, 90, 90, 91, 90, 92, 92, 93, 92, 94, 94, 95, 94, 96, 96, 97, 96, 98, 98, 99, 98, 100, 100, 101, 100, 102, 102, 103, 102, 104, 104, 105, 104, 106, 106, 107, 106, 108, 108, 109, 108, 106, 106, 107, 106, 108, 108, 109, 108, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 110, 110, 111, 110, 112, 112, 113, 112, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 94, 94, 95, 94, 100, 100, 101, 100, 122, 122, 123, 122, 124, 124, 125, 124, 90, 126, 91, 91, 92, 127, 93, 93, 94, 128, 95, 95, 96, 129, 97, 97, 98, 130, 99, 99, 100, 131, 101, 101, 102, 132, 103, 103, 104, 133, 105, 105, 106, 134, 107, 107, 108, 135, 109, 109, 106, 134, 107, 107, 108, 135, 109, 109, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 110, 136, 111, 111, 112, 137, 113, 113, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 94, 128, 95, 95, 100, 131, 101, 101, 122, 142, 123, 123, 124, 143, 125, 125, 90, 90, 91, 90, 92, 92, 93, 92, 94, 94, 95, 94, 96, 96, 97, 96, 98, 98, 99, 98, 100, 100, 101, 100, 102, 102, 103, 102, 104, 104, 105, 104, 106, 106, 107, 106, 108, 108, 109, 108, 106, 106, 107, 106, 108, 108, 109, 108, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 94, 94, 95, 94, 96, 96, 97, 96, 110, 110, 111, 110, 112, 112, 113, 112, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 114, 114, 115, 114, 116, 116, 117, 116, 118, 118, 119, 118, 120, 120, 121, 120, 94, 94, 95, 94, 100, 100, 101, 100, 122, 122, 123, 122, 124, 124, 125, 124, 90, 126, 91, 91, 92, 127, 93, 93, 94, 128, 95, 95, 96, 129, 97, 97, 98, 130, 99, 99, 100, 131, 101, 101, 102, 132, 103, 103, 104, 133, 105, 105, 106, 134, 107, 107, 108, 135, 109, 109, 106, 134, 107, 107, 108, 135, 109, 109, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 94, 128, 95, 95, 96, 129, 97, 97, 110, 136, 111, 111, 112, 137, 113, 113, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 114, 138, 115, 115, 116, 139, 117, 117, 118, 140, 119, 119, 120, 141, 121, 121, 94, 128, 95, 95, 100, 131, 101, 101, 122, 142, 123, 123, 124, 143, 125, 125, 90, 144, 91, 144, 92, 145, 93, 145, 94, 146, 99, 146, 96, 147, 101, 147, 98, 148, 99, 148, 100, 149, 101, 149, 102, 150, 103, 150, 104, 151, 105, 151, 106, 152, 107, 152, 108, 153, 109, 153, 106, 152, 107, 152, 108, 153, 109, 153, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 110, 154, 99, 154, 112, 155, 101, 155, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 94, 146, 99, 146, 100, 149, 101, 149, 122, 160, 99, 160, 124, 161, 101, 161, 90, 126, 91, 162, 92, 127, 93, 163, 94, 128, 99, 164, 96, 129, 101, 165, 98, 130, 99, 166, 100, 131, 101, 167, 102, 132, 103, 168, 104, 133, 105, 169, 106, 134, 107, 170, 108, 135, 109, 171, 106, 134, 107, 170, 108, 135, 109, 171, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 110, 136, 99, 172, 112, 137, 101, 173, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 94, 128, 99, 164, 100, 131, 101, 167, 122, 142, 99, 178, 124, 143, 101, 179, 90, 144, 91, 144, 92, 145, 93, 145, 94, 146, 99, 146, 96, 147, 101, 147, 98, 148, 99, 148, 100, 149, 101, 149, 102, 150, 103, 150, 104, 151, 105, 151, 106, 152, 107, 152, 108, 153, 109, 153, 106, 152, 107, 152, 108, 153, 109, 153, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 94, 146, 99, 146, 96, 147, 101, 147, 110, 154, 99, 154, 112, 155, 101, 155, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 114, 156, 99, 156, 116, 157, 101, 157, 118, 158, 99, 158, 120, 159, 101, 159, 94, 146, 99, 146, 100, 149, 101, 149, 122, 160, 99, 160, 124, 161, 101, 161, 90, 126, 91, 162, 92, 127, 93, 163, 94, 128, 99, 164, 96, 129, 101, 165, 98, 130, 99, 166, 100, 131, 101, 167, 102, 132, 103, 168, 104, 133, 105, 169, 106, 134, 107, 170, 108, 135, 109, 171, 106, 134, 107, 170, 108, 135, 109, 171, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 94, 128, 99, 164, 96, 129, 101, 165, 110, 136, 99, 172, 112, 137, 101, 173, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 114, 138, 99, 174, 116, 139, 101, 175, 118, 140, 99, 176, 120, 141, 101, 177, 94, 128, 99, 164, 100, 131, 101, 167, 122, 142, 99, 178, 124, 143, 101, 179
};
static const GraphemeSegmentationResult GraphemeSegmentationResult_t2[2880] = {
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=0, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=0, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=1, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=0, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=0, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_AtStart, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_None, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Prepend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_CR, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LF, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Control, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_Extend, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Regional_Indicator, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_SpacingMark, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_L, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_V, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_T, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LV, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_LVT, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
	{.grapheme_break=GBP_ZWJ, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=1},
	{.grapheme_break=GBP_Private_Expecting_RI, .incb_consonant_extended=1, .incb_consonant_extended_linker=0, .incb_consonant_extended_linker_extended=1, .emoji_modifier_sequence=1, .emoji_modifier_sequence_before_last_char=1, .add_to_current_cell=0},
};
static inline uint16_t GraphemeSegmentationKey(GraphemeSegmentationResult r, CharProps ch){
	return (r.state << 7) | ch.grapheme_segmentation_property;
}
// GraphemeSegmentationStateDeclaration: uses 9 bits {{{
typedef union GraphemeSegmentationState {
    struct __attribute__((packed)) {
#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
        uint8_t emoji_modifier_sequence_before_last_char : 1;
        uint8_t emoji_modifier_sequence : 1;
        uint8_t incb_consonant_extended_linker_extended : 1;
        uint8_t incb_consonant_extended_linker : 1;
        uint8_t incb_consonant_extended : 1;
        uint8_t grapheme_break : 4;
        uint8_t : 7;
#elif __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
        uint8_t : 7;
        uint8_t grapheme_break : 4;
        uint8_t incb_consonant_extended : 1;
        uint8_t incb_consonant_extended_linker : 1;
        uint8_t incb_consonant_extended_linker_extended : 1;
        uint8_t emoji_modifier_sequence : 1;
        uint8_t emoji_modifier_sequence_before_last_char : 1;
#else
#error "Unsupported endianness"
#endif
    };
    uint16_t val;
} GraphemeSegmentationState;
static_assert(sizeof(GraphemeSegmentationState) == sizeof(uint16_t), "Fix the ordering of GraphemeSegmentationState");
// EndGraphemeSegmentationStateDeclaration }}}

